import 'dart:async';

import 'package:rxdart/rxdart.dart' show RaceStream, BehaviorSubject;

/// the error returned by Context.err() when the context is canceled.
final exceptionCanceled = Exception("canceled");

/// the error returned by Context.err() when the context's deadline passes.
final exceptionDeadlineExceeded = Exception("deadline exceeded");

abstract class Context {
  static Context background() => _EmptyContext();

  static CancelableContext withCancel(Context parent) {
    return _CancelContext(parent);
  }

  static CancelableContext withTimeout(Context parent, Duration timeout) {
    return _TimerContext(parent, DateTime.now().add(timeout));
  }

  static CancelableContext withDeadline(Context parent, DateTime deadline) {
    return _TimerContext(parent, deadline);
  }

  static Context withValue(Context parent, dynamic key, dynamic val) {
    if (key != key) {
      throw Exception("$key, key must be equalable.");
    }
    return _ValueContext(parent, val: val, key: key);
  }

  static Context withTypedValue<T>(Context parent, T val) {
    return _ValueContext(parent, val: val);
  }

  /// returns valid DateTime when deadline is set.
  DateTime? deadline();

  /// returns a Stream that's closed when work done on behalf of this context should be canceled.
  Stream<Exception?>? done();

  /// If Context.done() is not yet closed, Err returns nil.
  /// If Context.done() is closed, Err returns a non-nil error explaining why:
  /// exceptionCanceled if the context was canceled
  /// exceptionDeadlineExceeded if the context's deadline passed.
  Exception? err();

  /// returns the value associated with this context for key or matched type
  ///
  /// ```dart
  /// var ctx = Context.withValue(
  ///     Context.background(),
  ///     "comparable-key",
  ///     "value will put into context",
  /// );
  ///
  /// ctx.value("comparable-key")
  /// ```
  ///
  /// ```dart
  /// class CustomType {
  /// }
  ///
  /// var ctx = Context.withTypedValue(
  ///     Context.background(),
  ///     CustomType(),
  /// );
  ///
  /// ctx.value<CustomType>();
  /// ```
  T? value<T>([dynamic key]);
}

abstract class CancelableContext extends Context {
  /// cancel context may with Exception
  ///
  /// the Exception notice to Stream's subscription from Context.done();
  /// the default Exception is exceptionCanceled.
  cancel(Exception? reason);
}

class _EmptyContext implements Context {
  @override
  Stream<Exception?>? done() {
    return null;
  }

  @override
  T? value<T>([key]) {
    return null;
  }

  @override
  DateTime? deadline() {
    return null;
  }

  @override
  Exception? err() {
    return null;
  }
}

class _ChildContext implements Context {
  Context? parent;

  @override
  Stream<Exception?>? done() {
    return parent?.done();
  }

  @override
  T? value<T>([key]) {
    return parent?.value<T>(key);
  }

  @override
  DateTime? deadline() {
    return parent?.deadline();
  }

  @override
  Exception? err() {
    return parent?.err();
  }
}

class _ValueContext with _ChildContext {
  dynamic key;
  dynamic val;

  _ValueContext(Context parent, {required this.val, this.key}) {
    this.parent = parent;
  }

  @override
  T? value<T>([k]) {
    if (k == null) {
      if (val is T) {
        return val;
      }
      return parent?.value<T>(k);
    }

    if (k == key && val is T) {
      return val;
    }
    return parent?.value<T>(k);
  }
}

class _CancelableContext implements CancelableContext {
  final BehaviorSubject<Exception?> _done = BehaviorSubject();
  Context? parent;
  Function()? _unsubscribe;

  @override
  cancel(Exception? reason) {
    _cancel(reason ?? exceptionCanceled);
    if (_unsubscribe != null) {
      _unsubscribe!();
    }
  }

  _cancel(Exception reason) {
    if (_done.valueOrNull != null) {
      return;
    }
    _done.sink.add(reason);
    _done.close();
  }

  @override
  Stream<Exception?>? done() {
    return _done.stream;
  }

  @override
  Exception? err() {
    return _done.valueOrNull ?? parent?.err();
  }

  @override
  DateTime? deadline() {
    return parent?.deadline();
  }

  @override
  T? value<T>([key]) {
    return parent?.value<T>(key);
  }
}

class _CancelContext with _CancelableContext {
  _CancelContext(Context parent) {
    this.parent = parent;
    _unsubscribe ??= _subscribeCancelsIfExists(parent, cancel: _cancel);
  }
}

class _TimerContext with _CancelableContext {
  _TimerContext(Context parent, this._deadline) {
    this.parent = parent;
    _unsubscribe ??= _subscribeCancelsIfExists(
      parent,
      cancel: _cancel,
      cancel$: Stream.periodic(_deadline.difference(DateTime.now()))
          .map((_) => exceptionDeadlineExceeded),
    );
  }

  final DateTime _deadline;

  @override
  DateTime? deadline() {
    return _deadline;
  }
}

_subscribeCancelsIfExists(
  Context parent, {
  required Function(Exception) cancel,
  Stream<Exception?>? cancel$,
}) {
  var parentCancel$ = parent.done();

  var sub = RaceStream<Exception?>([
    ...(parentCancel$ != null ? [parentCancel$] : []),
    ...(cancel$ != null ? [cancel$] : []),
  ]).listen(
    (reason) => reason != null ? cancel(reason) : null,
    onError: (err) => err != null ? cancel(err) : null,
  );

  return () => sub.cancel();
}
